"""
Commands for activating the test environment


"""
import sys
import os
import json
from zipfile import ZipFile
import multiprocessing
from livereload import Server
import click
import requests

from scintillant import addons_versions


def create_snlt():
    with open(os.getcwd() + '/.snlt', 'w+') as snlt:
        if sys.platform == 'win32':
            std_path = 'C:\\snlt\\testsuite\\'
        else:
            std_path = '/ust/testsuite/'
        path = input("Specify the path where Python can create "
                     f"the test environment files ({std_path}): ")
        path = std_path if not path else path
        click.echo(path)
        json.dump({
            "skill_name": "bottle-skill-template",
            "skill_working_url": "http://localhost:8080",
            "testsuite": {
                "version": addons_versions['__testsuite__'],
                "path": path,
                "theme": "dark"
            }
        }, snlt, sort_keys=True, indent=4)


def update_snlt():
    js = json.load(open(os.getcwd() + '/.snlt'))
    if sys.platform == 'win32':
        std_path = 'C:\\snlt\\testsuite\\'
    else:
        std_path = './testsuite/'
    path = input("Specify the path where Python can create "
                 f"the test environment files ({std_path}): ")
    path = std_path if not path else path
    with open(os.getcwd() + '/.snlt', 'w+') as snlt:
        if 'testsuite' in js and 'path' not in js['testsuite']:
            js['testsuite']['path'] = path
        elif 'testsuite' not in js:
            js['testsuite'] = {'theme': 'dark'}
            js['testsuite']['path'] = path
        json.dump(js, snlt, sort_keys=True, indent=4)


def install_suite(path):
    suite_release = ("https://github.com/PaperDevil/snlt_testsuite/releases/download/"
                     f"{addons_versions['__testsuite__']}"
                     "/dist.zip")
    r = requests.get(suite_release)
    os.makedirs(os.path.dirname(path + 'temp.zip'), exist_ok=True)
    with open(path + 'temp.zip', 'wb') as f:
        f.write(r.content)
    with ZipFile(path + 'temp.zip', 'r') as repo_zip:
        repo_zip.extractall(path)
    os.remove(path + 'temp.zip')


def get_current_app(app_path: str = None):
    try:
        click.echo("Tryng get application instance...")
        sys.path.append(app_path or os.getcwd())
        from manage import run_app
        return run_app
    except ImportError as exc:
        raise ImportError(
            "The specified path does not contain a valid manage.py file. "
            "Make sure you've installed all the dependencies from the "
            "requirements.txt file and created the run_app function correctly "
            "inside the manage.py file!"
        )


def start_bot_server():
    app = get_current_app()
    click.echo("Starting application server...")
    try:
        app(port=8080)
    except Exception as exc:
        click.echo(exc)


def start_test_suite():
    if os.path.exists(os.getcwd() + '/.snlt'):
        js = json.load(
            open(os.getcwd() + '/.snlt')
        )
        if 'testsuite' not in js or 'path' not in js['testsuite']:
            update_snlt()
            return start_test_suite()
        else:
            test_suite_path = js['testsuite']['path']
            install_suite(test_suite_path)
    else:
        create_snlt()
        return start_test_suite()

    if not os.path.exists(test_suite_path):
        install_suite(test_suite_path)
        return start_test_suite()

    click.echo("Creates processes...")
    _process = multiprocessing.Process(target=start_bot_server, args=())
    _process.start()

    click.echo("starting testsuite...")
    server = Server()
    server.serve(root=test_suite_path, default_filename='index.html')



"""
Commands aimed at generating skill templates.


"""
import os
from shutil import move, rmtree
from sys import version_info
from zipfile import ZipFile

import requests

from scintillant import BOTTLE_SKILL_TEMPLATE_URL


def start_bottle_skill(skill_name: str = None):
    """Download the latest skill template"""
    skill_name = input("Skill name: ") if not skill_name else skill_name
    skill_dirrectory = input(f"Skill directory ({skill_name}): ")
    if not skill_dirrectory:
        skill_dirrectory = skill_name
    print("Downloading project template...")

    r = requests.get(BOTTLE_SKILL_TEMPLATE_URL)
    with open('temp.zip', 'wb') as f:
        f.write(r.content)

    with ZipFile('temp.zip', 'r') as repo_zip:
        if skill_dirrectory == '.':
            repo_zip.extractall()
        else:
            repo_zip.extractall('.')

    os.remove('temp.zip')
    if not skill_dirrectory == '.':
        os.rename('bottle-skill-template-develop', skill_dirrectory)
    else:
        source = './bottle-skill-template-develop'
        files = os.listdir(source)
        for file in files:
            move(source + f'/{file}', os.getcwd())
        rmtree('./bottle-skill-template-develop')

    if not version_info < (3, 9):
        print("Language version below the required by the template")


def start_aiohttp_skill(skill_name: str = None):
    """Download the latest skill template
    TODO::
    """
    skill_name = input("Skill name: ") if not skill_name else skill_name
    print("Downloading project template...")
    #Repo.clone_from(AIOHTTP_SKILL_TEMPLATE_URL, getcwd() + f'/{skill_name}')


def start_fast_api_skill(skill_name: str = None):
    """Download the latest skill template
    TODO::
    """
    skill_name = input("Skill name: ") if not skill_name else skill_name
    print("Downloading project template...")
    #Repo.clone_from(FAST_API_SKILL_TEMPLATE_URL, getcwd() + f'/{skill_name}')


from abc import abstractmethod, ABC
from typing import Callable, Type
from pydantic import BaseModel

from scintillant.apimodels import SkillResponse
from scintillant.apimodels.types import SkillRequestType, SkillResponseType
from scintillant.outs import Dictionary

from loguru import logger


class Vault(BaseModel):

    @classmethod
    @abstractmethod
    def create(cls, data: SkillRequestType):
        pass


class ContextUpdater:
    states = {}  # state_name: state_func
    exit_phrases = ['exit', 'выход', 'хватит', 'остановись']
    vault = None

    def __init__(self, data: SkillRequestType,
                 dictionary: Type[Dictionary] = None,
                 vault: Type[Vault] = None):
        # Default parameters for find state and complete it
        self.data = data
        self.context = data.context
        self.response = SkillResponse(status='ok')

        if dictionary and data.user.client.type.lower() == 'voice':
            self.dictionary = dictionary.voice()
        elif dictionary:
            self.dictionary = dictionary.text()

        if vault and '__vault__' in self.context:
            self.vault = vault(**self.context['__vault__'])
        elif vault:
            self.vault = vault.create(data)
            self.context['__vault__'] = self.vault.dict()

    @property
    def next_state(self):
        if 'state' in self.context:
            return self.context['state']
        return None

    @next_state.setter
    def next_state(self, func: Callable):
        if func:
            self.context['state'] = func.__qualname__.split('.')[
                                        0] + '_' + func.__name__
        else:
            self.context.pop('state')

    def execute_state(self):
        # Termination of the skill if the user has asked for it
        if (self.data.update.in_text in self.exit_phrases
                or self.data.update.in_choice in self.exit_phrases):
            self.response.status = 'exit'
        # In case the user has just entered the skill,
        # we call the _initial_state_ function.
        elif 'state' not in self.context:
            self._initial_state_()
        # In case the user already has a state.
        else:
            state = self.next_state.split('ContextUpdater_', maxsplit=1)[1]
            if state not in self.states:
                raise Exception(f"State {state} not registered!")
            else:
                self.states[state](self)
        # Save vault updated data
        if self.vault:
            self.context['__vault__'] = self.vault.dict()

    def execute_now(self, func: Callable) -> None:
        self.next_state = func
        self.execute_state()

    def redirect(self, redirect_to: str) -> None:
        self.response.status = 'redirect'
        self.response.redirect_to = redirect_to

    def get_response(self) -> SkillResponseType:
        logger.debug(
            "Executing state {} from User {} | Update data: {}".format(
                self.next_state,
                self.data.user,
                self.data
            ))
        self.execute_state()
        self.response.context = self.context
        logger.info("Getting response: {}".format(self.response))
        if not self.response.out_text and self.response.status not in ['redirect', 'exit']:
            logger.error(
                'Response out_text field is None inside non-action state!')
        return self.response

    def _initial_state_(self):
        """Execute when state is None"""
        self.response.out_text = "Добро пожаловать в фреймворк Scintillant." \
                                 "\n\n" \
                                 "Вы сгенерировали ваш первый шаблон навыка " \
                                 "и видите это сообщение потому, что не " \
                                 "изменили ответ для шага `_initial_state_`!"

    @classmethod
    def statefunc(cls, func):
        cls.states[
            func.__qualname__.split('.')[0] + '_' + func.__name__] = func

        def wrapper(*args, **kwargs):
            func(*args, **kwargs)

        wrapper.__name__ = func.__qualname__.split('.')[
                               0] + '_' + func.__name__

        return wrapper

    def setup_logs(self, folder, level='DEBUG', rotation='2 GB'):
        logger.add(folder + f'{self.__class__}.log', level=level, rotation=rotation)
